#include<stdio.h>
#include<ctype.h>
#include<windows.h>
#include<string>
#include<stack>
#include"bit_stack.cpp"

class Output_interface;

//!  
class Options{
public:
	FILE *in_file;						//!<  
	FILE *out_file;						//!<  
	Output_interface *out_interface;	//!<  
public:
	Options();
	enum {none=0,ident,full}fmt_mode;
	int in_tab_size;					//!<    
	int out_tab_size;					//!<  
	char *FindArg(char**,int,char*);			//!<  
	void SetArg(char**,int,char*,int*,int,int);	//!<   int 
	void ScanOptions(char**,int);			//!<  
}options;


Options::Options()
{
	out_interface=0;
	in_tab_size=4;	//   
	out_tab_size=4;	// ...
}


//!  
class Output_interface{
protected:
	bool isKeyWord(char*);			//!<      
	virtual void PrintChar(char);	//!<  
public:
	virtual void Start() {}			//!<  
	virtual void Stop() {}			//!<  
	virtual void Ident(int);		//!< 
	virtual void Write(char*);		//!<  
	virtual void ScanOptions(char**,int) {} //!<  
};

//!  
class Format_engine
{
	enum TokType{t_do=0,t_else,t_for,t_if,t_while,t_case,t_class,t_struct,t_enum};	//  
	struct _frln{
		int ident;			//!<  
		int lines;			//!<     
		char *preview;		//!<  (    )
		std::string buffer;	//!<   
	}freadln; //!<    
	struct _fstmt{	
		bool wait_balance;	//!<   )    
	}fstmt; //!<  
	struct _v{
		enum {
			work,	//!<
			tadd,	//!<     
			pop,	//!<      ( })
			_if,	//!<  if     else
			_else	//!<  else  .  
		}status;	//!<  
		int curr_ident;							//!<  
		int temp_ident;							//!<  
		std::stack<int>ident_stack;				//!<  
	}fh;	//  
	struct _h{
		enum pre_en{
			empty,	//!< 
			none,	//!<  
			def,	//!< #define
			inc,	//!< #include 
			unk		//!<  
		}last_pre;	 //!<    
		enum {
			work,
			noline,		//!<  {}  
			noident,	//!<     (   )
			noident2	//!<       
		}status;	//!<  
		BitStack stack;
	}fv;
	struct _fout{	
		bool new_line;		//!<  
		int ident;			//!<   
	}foutput; //!<  
	int arr_count,arr_size;
	struct item_st{int ident; int lines; char *text;}*array; //!<    
	char *sym_table;	//!<   (  )
	int balance1;	//!<  
	int balance2;	//!<   
	void Clear();									//!<   
	void AddItem(int ident,char *text,int lines);	//!<     
	void FormatReadLn(int ident,char *text);		//!<     
	void FormatStmt(int ident,char*,int lines);		//!<  
	void FormatH();									//!<   
	void FormatV();									//!<  
	void Output();									//!< 
	TokType TokenType(char*);						//!<  
public:
	Format_engine();
	~Format_engine();
	void Format(int,char);						
	void Format(int,char*);						
	void Stop();									//!<  
};




//!  
class Syntax_engine{
	enum STATUS{
		ready,				//!<  
		work,				//!<   (  )
		str,str1,			//!< 
		ch,ch1,				//!< 
		num,				//!< 
		float1,				//!<  ( )
		float2,				//!<  (   )
		float3,				//!<  (  )
		float4,				//!<  ( e   +-)
		float5,				//!<  (  -  )
		rem,				//!<   
		rem_single,			//!<   
		rem_mul,rem_mul1,	//!<   
		tok,				//!< 
		prepro1,			//!<    
		prepro2,			//!<   
		space				//!<  
	}status;				//!<  
	int in_pos;				//!<    
	char old;				//!<  
	char sym;				//!<  
	int space_count;		//!<  
	std::string buffer;		//!<   
	bool FirstPair(char sym);			//!<       
	bool DetectPair(char old,char sym);	//!<     
	void CheckSymbol();					//!<   
protected:
	Format_engine form_engine;
	void Tabulator(char);		//!<   ,     
	void FormatBuffer();
	void FormatChar(char);
public:
	void Work(); //!<  
};


//! HTML 
class HTML:public Output_interface
{
	void PrintChar(char sym);	//!< 
	void Write(char*);			//!<  
	void Start();
	void Stop();
};

//! CPP 
class CPP:public Output_interface
{
	int no_tab;			//!<   
	int out_pos;		//!<     
public:
	CPP();
	void Ident(int);			//!< 
	void PrintChar(char);
	void ScanOptions(char**,int);	//!<  
	void Start();
};

//! TeX 
class TEX:public Output_interface
{
	void PrintChar(char sym);
public:
	void Write(char*);			//!<  
	void Start();
	void Stop();
};

//----- output Interface

//! \param text  
//! \result true   
bool Output_interface::isKeyWord(char *text)
{
	int a,b,c,n;
	static const char *list[]={
	"__int64","_asm","bool","break","case","char","class","const",
	"continue","default","do","double",
	"else","enum","extern","float","for","if","int","long",
	"public","private","protected","register","return","short",
	"signed","sizeof","static","struct","switch","template","typedef",
	"typename","union","unsigned","void","volatile","while"
	};
	a=0;
	b=sizeof(list)/sizeof(char*)-1;
	while(a<=b) {
#ifdef _DEBUG
		if(strcmp(list[a],list[b])>0) _asm { int 3 }
#endif
		c=(a+b)/2;
		n=strcmp(list[c],text);
		if(!n) return true;
		if(n>0) a=c+1; else b=c-1;
	}
	return false;
}

//! \param count  
void Output_interface::Ident(int count)
{
	while(count>0) fputc(' ',options.out_file),count--;
}

//! \param  
void Output_interface::Write(char* text)
{
	while(*text) PrintChar(*text++);
}

//! \param sym  
void Output_interface::PrintChar(char sym)
{
	fputc(sym,options.out_file);
}


//----- Format Engine

Format_engine::Format_engine()
{
	foutput.ident=0; 
	foutput.new_line=true; 
	fh.status=fh.work;
	fh.curr_ident=fh.temp_ident=0;
	fv.status=fv.work;
	fv.last_pre=fv.empty;
	freadln.preview="";
	fstmt.wait_balance=false;
	array=0; arr_size=arr_count=0;
	balance1=balance2=0;
	sym_table=(char*)malloc(512);
	for(int n=0; n<256; n++) sym_table[n*2]=n,sym_table[n*2+1]=0;
}


Format_engine::~Format_engine()
{
	free(array);
	free(sym_table);
}

//! \param ident   
//! \param text  
//! \param lines     
void Format_engine::AddItem(int ident,char *text,int lines)
{
	if(arr_count>=arr_size) {
		arr_size+=arr_size/2+1;
		array=(item_st*)realloc(array,arr_size*sizeof(item_st));
	}
	array[arr_count].ident=ident;
	array[arr_count].lines=lines;
	if(text[1]) array[arr_count].text=strdup(text);
	else array[arr_count].text=&sym_table[2*(byte)*text];
	arr_count++;
}


void Format_engine::Clear()
{
	while(arr_count) {
		arr_count--;
		if(array[arr_count].text[1]) free(array[arr_count].text);
	}
}


//! \param ident   
//! \param sym 
void Format_engine::Format(int ident,char sym)
{
	char buff[2];
	buff[0]=sym; buff[1]=0;
	Format(ident,buff);
}


//! \param ident   
//! \param text  
void Format_engine::Format(int ident,char *text)
{
	FormatReadLn(ident,text);
}


//! \param ident   
//! \param text  
void Format_engine::FormatReadLn(int ident,char *text)
{
	if(text[0]=='\n') freadln.lines++;
	else {
		freadln.preview=text;
		if(!freadln.buffer.empty()) {
			FormatStmt(freadln.ident,(char*)freadln.buffer.c_str(),freadln.lines);
		}
		freadln.lines=0;
		freadln.ident=ident;
		freadln.buffer=text;
	}
}

//! \param txt   
//! \result   
Format_engine::TokType Format_engine::TokenType(char *txt)
{
	int n;
	char *list[]={"do","else","for","if","while","case"
		,"class","struct","enum",0};
	for(n=0; list[n]; n++) if(!strcmp(txt,list[n])) break;
	return (TokType)n;
}

//!    .      .
void Format_engine::Stop()
{
	if(!freadln.buffer.empty()) {
		FormatStmt(freadln.ident,(char*)freadln.buffer.c_str(),freadln.lines);
	}
	if(arr_count) Output();
}


//! \param ident   
//! \param text  
//! \param nl     
void Format_engine::FormatStmt(int ident,char *text,int nl)
{
	switch(text[0]) {
	case '\n': return;
	case '(': balance1++; break;
	case ')': balance1--; break;
	case '{': Output(); balance2++; break;
	case '}': Output(); balance2--; break;
	}
	if(text[0]!='\n') AddItem(ident,text,nl);
	switch(text[0]) {
	case '/': if(text[1]=='/' || text[1]=='*') Output(); break;
	case ';':
	case ',': if(!balance1) Output(); break;
	case ':': if(balance2 && !text[1]) Output(); break;
	case ')': if(!balance1 && fstmt.wait_balance) Output(),fstmt.wait_balance=false; break;
	case '#':
	case '{':
	case '}': Output(); break;
	}
	switch(TokenType(text)) {
	case t_do: case t_else: Output(); break;
	case t_while: case t_for: case t_if: fstmt.wait_balance=true; break;
	}
}


void Format_engine::FormatH()
{
	TokType type=TokenType(array[0].text);
	switch(fh.status) {
	case fh._if:
	case fh.tadd:
		if(array[0].text[0]!='{') fh.temp_ident+=options.out_tab_size;
		break;
	case fh.pop:
		if(fh.ident_stack.empty()) break;
		fh.curr_ident=fh.ident_stack.top(),fh.ident_stack.pop();
		fh.temp_ident=fh.ident_stack.top(),fh.ident_stack.pop();
		if(fh.temp_ident) fh.status=fh._if; 
		else break;
	case fh._else: 
		if(type==t_else && fh.temp_ident) {fh.temp_ident-=options.out_tab_size;break;}
	default:
		fh.temp_ident=0;
		break;
	}
	foutput.ident=fh.curr_ident+fh.temp_ident;
	if(fh.status==fh._if) fh.status=fh._else;
	else fh.status=fh.work;
	if(arr_count>1 && foutput.ident && !strcmp(array[1].text,":") || type==t_case) foutput.ident-=options.out_tab_size;
	if(array[0].text[0]=='{') {
		fh.ident_stack.push(fh.status==fh._else?fh.temp_ident+options.out_tab_size:0);
		fh.ident_stack.push(fh.curr_ident);
		fh.curr_ident=foutput.ident+options.out_tab_size; 
		fh.status=fh.work;
	}
	if(array[0].text[0]=='}') {
		foutput.ident-=options.out_tab_size;
		fh.status=fh.pop;
	}
	switch(type) {
	case t_if:  fh.status=fh._if; break;
	case t_else: 
	case t_while: case t_for: fh.status=fh.tadd; break;
	}
}


void Format_engine::FormatV()
{
	int n; bool newline=true; bool dup=false;
	for(n=0; n<arr_count; n++) array[n].lines=0;	//     
	// #
	_h::pre_en old=fv.last_pre;
	if(array[0].text[0]!='#') fv.last_pre=fv.none;
	else if(!strncmp(array[0].text,"#define" ,7)) fv.last_pre=fv.def;
	else if(!strncmp(array[0].text,"#include",8)) fv.last_pre=fv.inc;
	else fv.last_pre=fv.unk;
	if(old!=fv.last_pre && old!=fv.empty) options.out_interface->Write("\n");
	// {}
	if(array[0].text[0]=='{') fv.stack.Push(fv.status!=fv.noline);
	if(array[0].text[0]=='}') {
		newline=fv.stack.Pop(); 
		dup=!balance2;
		if(freadln.preview[0]==';') newline=false;
	}
	switch(fv.status) {
	case fv.noident2: dup=true;
	case fv.noident:  array[0].ident=0;
	}
	fv.status=fv.work;
	// stmt
	switch(TokenType(array[0].text)) {
	case t_for: case t_if: case t_else: case t_do: case t_while: 
		if(freadln.preview[0]==';') newline=false;	//  ;   while      
		break;
	case t_class: case t_struct: case t_enum:
		fv.status=fv.noline;
		break;
	}
	n--;
	// \n
	if(newline) {
		array[n].lines=1;
		if(dup) array[n].lines++;
	}
	else {
		fv.status=dup?fv.noident2:fv.noident;
	}
}


void Format_engine::Output()
{
	int n,m;
	if(!arr_count) return;
	if(options.fmt_mode!=options.none) FormatH();
	if(options.fmt_mode==options.full) FormatV();
	for(n=0; n<arr_count; n++) {
		options.out_interface->Ident((foutput.new_line && options.fmt_mode!=options.none)?foutput.ident:array[n].ident);
		options.out_interface->Write(array[n].text);
		for(m=0; m<array[n].lines; m++) options.out_interface->Write("\n");
		foutput.new_line=!!m;
	}
	Clear();
}



//----- Syntax Engine

void Syntax_engine::Work()
{
	char sym;
	in_pos=0;
	old=0;
	space_count=0;
	status=ready;
	options.out_interface->Start();
	for(;;) {
		sym=fgetc(options.in_file); if(feof(options.in_file)) break;
		Tabulator(sym);
	}
	if(FirstPair(old)) FormatChar(old);
	FormatBuffer();
	form_engine.Stop();
}



//!       
void Syntax_engine::Tabulator(char s)
{
	sym=s;
	switch(s) {
	case '\t': sym=' '; do{CheckSymbol(),in_pos++;}while(in_pos%options.in_tab_size); break;
	case '\r': return;
	case '\n': CheckSymbol(); in_pos=0; break;
	default:   CheckSymbol(); in_pos++; break;
	}
	old=sym;
}

//! \param sym  
//! \result true       
bool Syntax_engine::FirstPair(char sym)
{
	return sym && strchr("<>=!+-:&|",sym)!=0;
}

//! \param sym   
//! \param old   
//! \param true    
bool Syntax_engine::DetectPair(char old,char sym)
{
	static const char ops[]=">><<<=>=->::++--==&&||!=";	//    
	for(int n=0; ops[n]; n+=2) if(ops[n]==old && ops[n+1]==sym) return true;
	return false;
}


void Syntax_engine::CheckSymbol()
{
l1:
		if(status==rem) {	//       
			if(sym=='/') {status=rem_single; buffer+="//";return;}
			else if(sym=='*') {status=rem_mul; buffer+="/*"; return;}
			else FormatChar('/'),status=work;
		}
		switch(status) {
		case ready:		//  ,   
			if(sym=='#') {status=prepro1; buffer+=sym;break;}
		case work:		//    
			if(DetectPair(old,sym)) {buffer+=old,buffer+=sym,sym=0,FormatBuffer();return;}
			if(FirstPair(old)) FormatChar(old);
			if(FirstPair(sym)) return;
			if(sym=='\n') break;
			switch(sym) {
			case ' ':  status=space; space_count=0; break;
			case '"':  status=str; break;
			case '\'': status=ch; break;
			case '.':  status=float1; break;
			case '/':  status=rem; break;
			default:
				if(isdigit(sym)) status=num;
				else if(isalpha(sym) || sym=='_') status=tok;
				else status=work,FormatChar(sym);
			}
			break;
			case prepro1:	//     (   #
				if(sym!=' ') status=prepro2; //else UpdateIdent(sym);
				break;
		}
		if(sym=='\n' && status!=rem_mul && status!=rem_mul1) { //     
			FormatBuffer(); 
			FormatChar('\n');
			status=ready; 
		} else {
			bool fl=false;
			switch(status) {	//     
			case space: if(sym!=' ') fl=true; break;
			case tok: if(!isalnum(sym) && sym!='_') fl=true; break;
			case float4:
				status=float5;
				if(sym=='+' || sym=='-') break;
			case float3:
			case num: 
				if(sym=='e' || sym=='E') {status=float4;break;}
				if(sym=='.') {status=float2; break;}
			case float2:
			case float5: if(status==float2) status=float3;
				if(!isdigit(sym)) fl=true; 
				break;
			case float1: status=float2; break;
			}
			if(fl) {FormatBuffer(),status=work;goto l1;}
			switch(status) {		//   
			case space: space_count++; break;
			case tok:
			case num:
			case float2: case float3: case float4: case float5:
			case rem_single: case rem_mul: case rem_mul1:
			case str: case str1:
			case ch: case ch1:
			case prepro2:
				buffer+=sym;
				break;
			}
			switch(status) {	//     
			case rem_mul: if(sym=='*') status=rem_mul1; break;
			case rem_mul1: if(sym=='/') FormatBuffer(),status=work; else status=rem_mul; break;
			case str1: status=str; break;
			case str:
				if(sym=='\\') status=str1;
				if(sym=='"' && buffer.size()>1) FormatBuffer(),status=work;
				break;
			case ch1: status=ch; break;
			case ch:
				if(sym=='\\') status=ch1;
				if(sym=='\'' && buffer.size()>1) FormatBuffer(),status=work;
				break;
			}
		}
}

//!     
void Syntax_engine::FormatBuffer()
{
	if(buffer.empty()) return;
	form_engine.Format(space_count,(char*)buffer.c_str());
	buffer.erase();
	space_count=0;
}


void Syntax_engine::FormatChar(char sym)
{
	form_engine.Format(space_count,sym);
	space_count=0;
}




//----- HTML

void HTML::PrintChar(char sym)
{
	switch(sym) {
	case '<': fprintf(options.out_file,"&lt;");  break;
	case '>': fprintf(options.out_file,"&gt;");  break;
	case '&': fprintf(options.out_file,"&amp;"); break;
	default:  fputc(sym,options.out_file);       break;
	}
}


void HTML::Write(char *text)
{
	char *cls=0; 
	char sym=*text;
	if(sym=='#') cls="prepro"; else
	if(sym=='"' || sym=='\'') cls="str"; else
	if(sym=='/' && (text[1]=='/' || text[1]=='*')) cls="rem"; else
	if(isKeyWord(text)) cls="key";
	if(cls) printf("<span class='%s'>",cls);
	while(*text) PrintChar(*text++);
	if(cls) printf("</span>");
}

void HTML::Start()
{
	fprintf(options.out_file,
		"<html><style type='text/css'><!--\n"
		".key{color:blue}\n"
		".str{color:orange}\n"
		".rem{color:green}\n"
		".prepro{color:red}\n"
		"--></style><body><pre>\n"
		);
}


void HTML::Stop()
{
	fprintf(options.out_file,"\n</pre></body></html>\n");
}

//----- cpp

CPP::CPP()
{
	no_tab=false;
}


void CPP::ScanOptions(char** list,int count)
{
	if(options.FindArg(list,count,"-no-tab")) no_tab=true;
}


void CPP::Start()
{
	out_pos=0;
}

void CPP::PrintChar(char sym)
{
	Output_interface::PrintChar(sym);
	out_pos++;
	if(sym=='\n') out_pos=0;
}

void CPP::Ident(int count)
{
	int n;
	if(!no_tab && count>1) while((out_pos+count)/options.out_tab_size>out_pos/options.out_tab_size) {
		putchar('\t');
		n=options.out_tab_size-out_pos%options.out_tab_size;
		count-=n;
		out_pos+=n;		
	}
	Output_interface::Ident(count);
	out_pos+=count;
}

//----- TeX

void TEX::Start()
{
	fprintf(options.out_file,
		"\\documentclass{article}\\usepackage{alltt}\\oddsidemargin=-16pt"
		"\\begin{document}\n"
		"\\catcode`<=11\\catcode`>=11\n"
		"\\tt\\begin{alltt}\n"
		);
}

void TEX::Stop()
{
	fprintf(options.out_file,
		"\\end{alltt}\n"
		"\\end{document}\n"
		);
}


void TEX::Write(char *text)
{
	bool key=isKeyWord(text);
	if(key) fprintf(options.out_file,"\\underline{");
	while(*text) PrintChar(*text++);
	if(key) fputc('}',options.out_file);
}


void TEX::PrintChar(char sym)
{
	switch(sym) {
	case '_':  fprintf(options.out_file,"\\underline{ }"); break;
	case '{':  fprintf(options.out_file,"\\{"); break;
	case '}':  fprintf(options.out_file,"\\}"); break;
	case '\\': fprintf(options.out_file,"\\(\\backslash\\)"); break;
	default: fputc(sym,options.out_file); break;
	}
}

//----- undef

//!       
//! \param list     
//! \param count  
//! \param key     
char *Options::FindArg(char **list,int count,char *key)
{
	int n,len; char *ref;
	len=strlen(key);
	for(n=1; n<count; n++) if(!strncmp(list[n],key,len)) {
		ref=&list[n][len];
		if(*ref==' ' || !*ref) return "";
		if(*ref=='=') return ref+1;
	}
	return 0;
}

//!       .
//!     ,  .
//! \param list     
//! \param count  
//! \param key     
//! \param par      
//! \param min    
//! \param max    
void Options::SetArg(char **list,int count,char *key,int *par,int min,int max)
{
	char *ref=FindArg(list,count,key);
	if(!ref || !*ref) return;
	*par=atoi(ref);
}


void Help()
{
	fprintf(stderr,
		"-help\n"
		"-out=[cpp,html,tex] *\n"
		"-fmt=[none,ident,full]\n"
		"-in-tab-size=<n>\n"
		"-out-tab-size=<n>\n"
		"-cpp-notab\n"
		);
}


void Options::ScanOptions(char **list,int count)
{
	char *ref;
	ref=options.FindArg(list,count,"-out"); if(ref) {
		if(!strncmp(ref, "cpp",1)) options.out_interface=new CPP;
		if(!strncmp(ref, "tex",2)) options.out_interface=new TEX;
		if(!strncmp(ref,"html",3)) options.out_interface=new HTML;
	}
	ref=options.FindArg(list,count,"-fmt"); if(ref) {
		if(!strncmp(ref, "none",4)) options.fmt_mode=none;
		if(!strncmp(ref,"ident",5)) options.fmt_mode=ident;
		if(!strncmp(ref, "full",4)) options.fmt_mode=full;
	}
	options.SetArg(list,count,"-in-tab-size",&options.in_tab_size,1,32);
	options.SetArg(list,count,"-oun-tab-size",&options.out_tab_size,1,32);
}


void main(int argc,char **argl)
{
	int mode=0; 
	if(options.FindArg(argl,argc,"-help")) {Help();return;}
	options.ScanOptions(argl,argc);
	if(!options.out_interface) {Help();return;}
	options.in_file=stdin;
	options.out_file=stdout;
	options.out_interface->ScanOptions(argl,argc);
	Syntax_engine engine;
	engine.Work();
}


